function init()

	self.drainPos = {};
	self.drainPos[0] = entity.position();
	self.drainPos[1] = {(self.drainPos[0][1])-1,self.drainPos[0][2]}
	self.drainPos[2] = {(self.drainPos[0][1])+1,self.drainPos[0][2]}

	-- start with the object position
	self.leftPumpPos = {self.drainPos[0][1]-2, self.drainPos[0][2]+3}
	self.rightPumpPos = {self.drainPos[0][1]+2, self.drainPos[0][2]}
	
	if (storage.liquidStored == nil) then
		storage.liquidStored = 0
		storage.liquidType = 0
	end
	self.reservoirSize = entity.configParameter("resevoirSize")
	self.containerSize = entity.configParameter("slotCount")
	self.containerFull = false
	
	self.liquidNameToLiquidId = {      
        liquidwater = 1,
        liquidlava = 2,
        liquidpoison = 3,
        liquidoil = 5,
        liquidhealing = 6,
        liquidmilk = 7,
        liquidfuel = 11,
        swampwater = 12,
        liquidslime = 13
    }
	
	self.liquidIdToLiquidName = {
		[1] = "liquidwater",
		[2] = "liquidlava",
		[3] = "liquidpoison",
		[5] = "liquidoil",
		[6] = "liquidhealing",
		[7] = "liquidmilk",
		[11] = "liquidfuel",
		[12] = "swampwater",
		[13] = "liquidslime"
	}

	self.storeToContainerBreakpoint = 100
	
	output(storage.liquidType)

end

function getMaxStorage()
	local reservoirSize = entity.configParameter("reservoirSize")
	if (self.liquidIdToLiquidName[storage.liquidType]) then
		reservoirSize = reservoirSize - (world.containerSize(entity.id())*1000)
	end
	return reservoirSize
end

-- Change Animation
function output(liquidType)
	if (not liquidType) or liquidType == 0 then
		entity.setAnimationState("sensorState", "off")
	elseif liquidType == 1 or liquidType == 6 then
		entity.setAnimationState("sensorState", "water")
	elseif liquidType == 3 then
		entity.setAnimationState("sensorState", "poison")
	elseif liquidType == 2 or liquidType == 8 then
		entity.setAnimationState("sensorState", "lava")
	elseif liquidType == 11 then
		entity.setAnimationState("sensorState", "fuel")
	else
		entity.setAnimationState("sensorState", "other")
	end
end

function destroyLiquid(i)
	world.forceDestroyLiquid(self.drainPos[i])
end

function addToReserveFromContainer(amountToAdd)
--	print("trying to add",storage.liquidType, "  from container")
--	print(self.liquidIdToLiquidName[storage.liquidType])

	local liquidName = self.liquidIdToLiquidName[storage.liquidType]
	if (liquidName) then
--		print("searching for liquid in container")
        local targetLiquid = {name=liquidName, count=amountToAdd}

        local liquidConsumed = world.containerConsume(entity.id(), targetLiquid)
        if (liquidConsumed) then
--			print ("liquid consumed from container")
            storage.liquidStored = storage.liquidStored + amountToAdd;
			self.containerFull = false
            return true
		else
--			print ("Unable to consume liquid")
			--Check if container has less than the breakpoint amout of liquid left
			--If so use whatever is left
			local contents = world.containerItems(entity.id())
			for i,item in pairs(contents) do
				if (item.name == liquidName) then
--					print("found small stack, using it")
					liquidConsumed = world.containerConsume(entity.id(), item)
					if (liquidConsumed) then
--						print ("consumed",item.count,"liquid")
						storage.liquidStored = storage.liquidStored + item.count
						self.containerFull = false
					end
				end
			end
            return false
		end
	else
		--Unknown liquid in resevoir
		return false
	end
end

function saveReserveToContainer()
--	print("trying to save resevour to container")
	local amountToSave = math.floor(storage.liquidStored);
	local liquidName = self.liquidIdToLiquidName[storage.liquidType]
	if (liquidName) then
--		print("liquid in resevour of known item type")
        local liquid = {name=liquidName, count=amountToSave}
        if (amountToSave > 0) and liquid then
           local overflow =  world.containerAddItems(entity.id(), liquid)
		   local overflowCount = 0
			if (overflow) then
				overflowCount = overflow.count
				self.containerFull = true
--				print("container overflowed with", overflowCount)
			end
			storage.liquidStored = storage.liquidStored - (amountToSave - overflowCount)
		end
	end
end

function pumpLiquid(inflowActive, outFlowActive)
-- if there is any liquid,
	local spawnAmount = 0;
	local isInvalidLiquid = false;
	local isOverflowing = false;

	if (inflowActive) then
        for i, drainPos in pairs(self.drainPos) do
            local liquidScan = world.liquidAt(drainPos)
            if (liquidScan) then
                local liquidType = liquidScan[1];
                local liquidDepth = liquidScan[2];

                --Aquire liquid
                if not(isReserveTypeSet()) then
                    storage.liquidType = liquidType
                    storage.liquidStored = liquidDepth
                    output(liquidType)
                    destroyLiquid(i)
                elseif (isInReserve(liquidType)) then
                    -- If will overflow and pumping save overflow to pump out
                    if ((storage.liquidStored + liquidDepth) > getMaxStorage()) then
                        if (outFlowActive) then
                            isOverflowing = true;
                            storage.liquidStored = storage.liquidStored + liquidDepth;
                            destroyLiquid(i)
                        else
                            --Don't store or destroy liquid
                        end
                    else
                        storage.liquidStored = storage.liquidStored + liquidDepth;
                        destroyLiquid(i)
                    end
                else
                    --Don't accept liquid of wrong type
                    isInvalidLiquid = true;
                end

                --If storage capacity reached then spawn all water else spawn half a tile
                if not (isInvalidLiquid) then
                    if (isOverflowing) then
                        spawnAmount = spawnAmount + liquidDepth;
                    else
                        if (liquidDepth > 0.5) then
                            spawnAmount = spawnAmount + 0.5;
                        else
                            spawnAmount = liquidDepth + storage.liquidStored
                        end
					end
				end
            end
        end
	end
	
	if (spawnAmount <= 0) and (isReserveTypeSet()) then
		spawnAmount = 1.5

		if (storage.liquidStored < 1.5) then
			spawnAmount = storage.liquidStored
		end
	end

	return spawnAmount;
end

-- Removes Liquids at current position
function pump(inflowActive, leftOutflowActive, rightOutflowActive)


	local spawnAmount = pumpLiquid (inflowActive, (leftOutflowActive or rightOutflowActive))

	if (spawnAmount < 1.5 and (leftOutflowActive or rightOutflowActive)) and
			addToReserveFromContainer(self.storeToContainerBreakpoint) then
		spawnAmount = 1.5
	end
	--If there is something to spawn and an output pump is on
	if (spawnAmount > 0) and (leftOutflowActive or rightOutflowActive) then
		--If both on then spawn half on each
		if (leftOutflowActive and rightOutflowActive) then
			spawnAmount = spawnAmount/2;
		end
		
		-- spawn the liquid
		if (leftOutflowActive)	then
			world.spawnLiquid(self.leftPumpPos, storage.liquidType, spawnAmount)
		end
		
		if(rightOutflowActive) then
			world.spawnLiquid(self.rightPumpPos, storage.liquidType, spawnAmount)
		end
		
		storage.liquidStored = (storage.liquidStored - spawnAmount);
		
		if (storage.liquidStored <= 0) then
			--If the pump ran completly dry pull from the container
			if (not addToReserveFromContainer(self.storeToContainerBreakpoint)) then
				--If there was nothing to add, display empty
				storage.liquidType = 0
				storage.liquidStored = 0
				output(0)
			end
		end
	end

	if (not self.containerFull) and (storage.liquidStored > self.storeToContainerBreakpoint) then
		saveReserveToContainer()
	end
	
end

function onInteraction(args)
	local liquidPrintout = "Liquid stored: " .. math.floor(storage.liquidStored)
	if (self.liquidIdToLiquidName[storage.liquidType]) then
		liquidPrintout = liquidPrintout .. " " .. self.liquidIdToLiquidName[storage.liquidType]
	else
		liquidPrintout = liquidPrintout .. " liquidId " .. storage.liquidType
	end
	entity.say(liquidPrintout)
	print(liquidPrintout)
end

function update(dt)
  local powered = entity.getInboundNodeLevel(0)
  local leftPump = entity.getInboundNodeLevel(1)
  local rightPump = entity.getInboundNodeLevel(2)

  
  pump(powered, leftPump, rightPump)
  
end

--Check for Match of liquid in container to liquid in reserve
function isInReserve(liquidId)
	if (liquidId) then
--		print("known liquid found of type:",liquidId)
		if (not(isReserveTypeSet())) or (liquidId == storage.liquidType) then
			--liquid matched storage or there is no liquid in the reserve
			return true
		else
			return false
		end
	end
	--unknown liquid 
	return false
end

--Removes the item from the continer and spawns it in the world
function ejectItem(item)
	if (item.count > 0) then
        world.containerConsume(entity.id(), item)
        local rejectPosition = entity.position()
        rejectPosition[2] = rejectPosition[2] + 7
        world.spawnItem(item.name, rejectPosition, item.count, {})
	end
end

--Utility method to check if the liquid type in the storage is set
function isReserveTypeSet()
	return (storage.liquidType > 0)
end

function onContainerCallback(args)
  local contents = world.containerItems(entity.id())
    local itemCount = 0
  	local rejectionList = {}
  	local rejectCount = 0
	for i, item in pairs (contents) do
		itemCount = itemCount + 1
		local itemName = item.name
		local liquidId = self.liquidNameToLiquidId[itemName]
		if (isInReserve(liquidId)) then
--			print ("liquid matches reserve")
			if (isReserveTypeSet()) then
--				print("reserve not empty, adding matching chest liquid to reserve")
			else
--				print("reserve empty, initing with container liquid")
				storage.liquidType = liquidId
				output(liquidId)
			end

		else
			rejectionList[i] = item
			rejectCount = rejectCount + 1
		end
	end

   --this checks the edge case that the last of the current liquid was removed by swapping it
   --with another liquid
  	if (itemCount == 1) and (rejectCount == 1) then
		for i, reject in pairs (rejectionList) do
			local liquidId = self.liquidNameToLiquidId[reject.name]
			if (liquidId) then
				local currentLiquidName = self.liquidIdToLiquidName[storage.liquidType]
				print("currentLiquidName: ", currentLiquidName)
				--If current liquid is known kick it out
				if (currentLiquidName) then
					storage.liquidType = liquidId
					local currentLiquid = {name=currentLiquidName, count=storage.liquidStored}
					ejectItem(currentLiquid)
					storage.liquidStored = 0
					output(liquidId)
				--If current liquid is not known but is less than 2 (the pump in use will have about that much)
				--then just destroy it
				elseif (storage.liquidStored < 2) then
					storage.liquidType = liquidId
					storage.liquidStored = 0
					output(liquidId)
				end
			else
				ejectItem(reject)
			end
		end
	else
		for i, reject in pairs (rejectionList) do
			ejectItem(reject)
		end
	end

    if (self.containerFull) then
		local currentLiquidName = self.liquidIdToLiquidName[storage.liquidType]
		if (currentLiquidName) then
			local tempLiquid = {name=currentLiquidName, count=1 }
			if (world.containerAvailable(entity.id(), tempLiquid) > 0)then
				self.containerFull = false
			end
		end
	end

	if (itemCount == 0) and (storage.liquidStored <1)  then
		storage.liquidType = 0
		output(0)
	end
end

function die()
	if (storage.liquidType > 0) and (storage.liquidStored > 0) then
		local liquidName = self.liquidIdToLiquidName[storage.liquidType]
		if (liquidName) then
			world.spawnItem(liquidName, entity.position(), storage.liquidStored, {})
		else
			world.spawnLiquid(entity.position(), storage.liquidType, storage.liquidStored);
		end
	end
end